// Copyright (C) 2001 Eric Hanchrow
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Eric Hanchrow -- offby1@blarg.net
// 1914 N. 40th
// Seatlle, WA, USA
// 98103-8310
#include <strstream>
#include <string>

#include <cstdlib>

// all of these are for the redirector class
#include <cerrno>
#include <cstdio>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>

#include <cppunit/TestCase.h>
#include <cppunit/TestCaller.h>
#include <cppunit/TestSuite.h>
#include <cppunit/TestResult.h>
#include <cppunit/TextTestResult.h>

#include "../card.hh"
#include "../auction.hh"
#include "../exceptions.hh"

#include "ugly_macros.hh"

#include "score-tests.hh"

using namespace CppUnit;

namespace {
class CardTest : public TestCase
{
private:
  card *_m_two_clubs;
  card *_m_ten_diamonds;
  card *_m_ace_spades;
public:
  void setUp ()
  {
    _m_two_clubs    = new card (0);
    _m_ten_diamonds = new card (21);
    _m_ace_spades   = new card (51);
  }

  void tearDown ()
  {
    delete _m_two_clubs;
    delete _m_ten_diamonds;
    delete _m_ace_spades;
  }

  // I don't understand why I need to define this.
  CardTest () : TestCase ("weird default") {}

  CardTest (std::string name) : TestCase (name) {}

  void testIllegal ()
  {
    CHECK_FOR_EXCEPTION (Bad_Card, ({card *_m_illegal = new card (52); delete _m_illegal;}))
  }

  void testInit ()
  {
    CPPUNIT_ASSERT (_m_two_clubs->get_rank () == two);
    CPPUNIT_ASSERT (_m_two_clubs->get_suit () == clubs);

    CPPUNIT_ASSERT (_m_ten_diamonds->get_rank () == ten);
    CPPUNIT_ASSERT (_m_ten_diamonds->get_suit () == diamonds);

    CPPUNIT_ASSERT (_m_ace_spades->get_rank () == ace);
    CPPUNIT_ASSERT (_m_ace_spades->get_suit () == spades);
  }

  void test_HCP ()
  {
    CPPUNIT_ASSERT (_m_two_clubs   ->get_HCP () == 0);
    CPPUNIT_ASSERT (_m_ten_diamonds->get_HCP () == 0);
    CPPUNIT_ASSERT (_m_ace_spades  ->get_HCP () == 4);
  }

private:
  void display_helper (const card *pC, const std::string &expected)
  {
    std::ostrstream actual_stream;
    actual_stream   << *pC << '\0';
    const std::string actual (actual_stream.str ());

    std::ostrstream message;
    message << "Expected `"
            << expected
            << "'; got `"
            << actual
            << "'";
    CPPUNIT_ASSERT_MESSAGE (message.str(),
                            actual == expected);
    message.freeze (0);
    actual_stream.freeze (0);
  }

public:
  void test_display ()
  {
    display_helper (_m_two_clubs   , "two of clubs");
    display_helper (_m_ten_diamonds, "ten of diamonds");
    display_helper (_m_ace_spades  , "ace of spades");
  }

  static Test *suite ()
  {
    TestSuite *suiteOfTests = new TestSuite;
    suiteOfTests->addTest (new TestCaller<CardTest>("init",             &CardTest::testInit));
    suiteOfTests->addTest (new TestCaller<CardTest>("high-card-points", &CardTest::test_HCP));
    suiteOfTests->addTest (new TestCaller<CardTest>("display",          &CardTest::test_display));
    suiteOfTests->addTest (new TestCaller<CardTest>("illegal-card",     &CardTest::testIllegal));
    return suiteOfTests;
  }
};

class seat_test: public TestCase
{
public:
  seat_test () : TestCase () {}
  seat_test (std::string name) : TestCase (name) {}

  void on_same_team_test ()
  {
    seat N;
    seat E (N + 1);
    seat S (E + 1);
    seat W (S + 1);
    CPPUNIT_ASSERT (N.on_same_team (S));
    CPPUNIT_ASSERT (S.on_same_team (N));
    CPPUNIT_ASSERT (E.on_same_team (W));
    CPPUNIT_ASSERT (W.on_same_team (E));
    CPPUNIT_ASSERT (!N.on_same_team (E));
    CPPUNIT_ASSERT (!N.on_same_team (W));
    CPPUNIT_ASSERT (!S.on_same_team (E));
    CPPUNIT_ASSERT (!S.on_same_team (W));
  }

  static Test *suite ()
  {
    TestSuite *suiteOfTests = new TestSuite;
    suiteOfTests->addTest (new TestCaller<seat_test>("on same team", &seat_test::on_same_team_test));
    return suiteOfTests;
  }
};

class auction_test : public TestCase
{
public:
  auction_test () : TestCase () {}
  auction_test (std::string name) : TestCase (name) {}
  void setUp ()
  {
  }

  void tearDown ()
  {
  }

  void testInit ()
  {
    auction a;

    CPPUNIT_ASSERT (!a._f_someone_has_opened ());
    CPPUNIT_ASSERT (a.get_state () == auction::in_progress);

    CHECK_FOR_EXCEPTION(Bad_State, a.get_contract ());

    CPPUNIT_ASSERT (0 == a.highest ());
    CPPUNIT_ASSERT (false == a.ns_vulnerable);
    CPPUNIT_ASSERT (false == a.ew_vulnerable);

    CPPUNIT_ASSERT (0 == a.size ());
    {
      std::vector<const call *> calls (a.calls ());
      CPPUNIT_ASSERT (0 == calls.size ());
    }
  }

  void test_opens_and_vulnerability ()
  {
    auction a;

    a.push_back (new pass ());

    CPPUNIT_ASSERT (1 == a.size ());
    {
      std::vector<const call *> calls (a.calls ());
      CPPUNIT_ASSERT (1 == calls.size ());
      CPPUNIT_ASSERT (dynamic_cast<const pass *>(calls.back ()));
    }

    a.push_back (new bid (1, spades));

    CPPUNIT_ASSERT (a._f_someone_has_opened ());
    CPPUNIT_ASSERT (a.get_state () == auction::in_progress);
    CHECK_FOR_EXCEPTION(Bad_State, a.get_contract ());

    CPPUNIT_ASSERT (0 != a.highest ());
    CPPUNIT_ASSERT (false == a.ns_vulnerable);
    CPPUNIT_ASSERT (false == a.ew_vulnerable);

    a.push_back (new pass ());
    CPPUNIT_ASSERT (a._f_someone_has_opened ());

  }

  void test_settled_contract ()
  {
    auction a;
    a.push_back (new bid (1, spades));
    a.push_back (new pass ());
    a.push_back (new pass ());
    CHECK_FOR_EXCEPTION (Bad_State, a.get_contract ());
    a.push_back (new pass ());
    CPPUNIT_ASSERT (a._f_someone_has_opened ());
    CPPUNIT_ASSERT (a.get_state () == auction::contract_settled);

    {
      const contract &c = a.get_contract ();
      const bid *pb = a.highest ();
      CPPUNIT_ASSERT (c.get_bid () == *pb);
      CPPUNIT_ASSERT (pb->level () == 1);
      CPPUNIT_ASSERT (pb->denom () == spades);
    }
  }

  // Doubles.

  /*
   * A player may double only the last preceding bid, and then only
   * if it was made by an opponent and no calls other than pass have
   * intervened.
   */
  /*
   * A player may redouble only the last preceding double, and then
   * only if it was made by an opponent and no calls other than pass
   * have intervened.
   */

  void cant_double_when_no_bids ()
  {
    auction a;
    CHECK_FOR_EXCEPTION(Bad_Double, a.push_back (new Double  (false)));
    CHECK_FOR_EXCEPTION(Bad_Double, a.push_back (new Double  (true)));

    a.push_back (new pass ());

    CHECK_FOR_EXCEPTION (Bad_Double, a.push_back (new Double (false)));
    CHECK_FOR_EXCEPTION (Bad_Double, a.push_back (new Double (true)));

  }

  void cant_double_partners_bid ()
  {
    auction a;
    a.push_back (new bid (1, spades));
    a.push_back (new pass ());
    CHECK_FOR_EXCEPTION (Bad_Double, a.push_back (new Double (false)));
    CHECK_FOR_EXCEPTION (Bad_Double, a.push_back (new Double (true)));
  }

  void may_double_opponent ()
  {
    {
      auction a;
      a.push_back (new bid (1, spades));
      CHECK_FOR_NO_EXCEPTION(a.push_back (new Double (false)));
    }

    // Similar to the above, but with two passes between the bid and
    // the double.
    {
      auction b;
      b.push_back (new bid (1, spades));
      b.push_back (new pass ());
      b.push_back (new pass ());
      CHECK_FOR_NO_EXCEPTION(b.push_back (new Double (false)));
    }
  }

  void cant_double_partners_double ()
  {
    auction a;
    a.push_back (new bid (1, spades));
    a.push_back (new Double (false));
    a.push_back (new pass ());
    CHECK_FOR_EXCEPTION(Bad_Double, a.push_back (new Double (false)));
    CHECK_FOR_EXCEPTION(Bad_Double, a.push_back (new Double (true)));
  }

  void cant_double_opponents_double ()
  {
    auction a;
    a.push_back (new bid (1, spades));
    a.push_back (new Double (false));
    CHECK_FOR_EXCEPTION (Bad_Double, a.push_back (new Double (false)));
  }

  void may_redouble_opponents_double ()
  {
    {
      auction a;
      a.push_back (new bid (1, spades));
      a.push_back (new Double (false));

      CHECK_FOR_NO_EXCEPTION (a.push_back (new Double (true)));
    }

    {
      auction b;
      b.push_back (new bid (1, spades));
      b.push_back (new Double (false));
      b.push_back (new pass ());
      b.push_back (new pass ());
      CHECK_FOR_NO_EXCEPTION (b.push_back (new Double (true)));
    }
  }

  static Test *suite ()
  {
    TestSuite *suiteOfTests = new TestSuite;
#define ADD_TEST(MEMBER_NAME) \
    suiteOfTests->addTest (new TestCaller<auction_test>((#MEMBER_NAME), &auction_test::MEMBER_NAME));

    ADD_TEST(cant_double_opponents_double);
    ADD_TEST(cant_double_partners_bid);
    ADD_TEST(cant_double_partners_double);
    ADD_TEST(may_double_opponent);
    ADD_TEST(may_redouble_opponents_double);
    ADD_TEST(testInit);
    ADD_TEST(test_opens_and_vulnerability);

    return suiteOfTests;
  }
};

class redirector
{
private:
  int _old_fd;
  int _fd_with_saved_info;
  int _new_fd;
public:
redirector (int old_fd,
            const std::string &fn)
  : _old_fd (old_fd),
    _fd_with_saved_info (-1),
    _new_fd (-1)
{
  _fd_with_saved_info = ::dup (_old_fd);

  if (-1 == _fd_with_saved_info)
    {
      throw ::strerror (errno);
    }

  _new_fd = ::open(fn.c_str(), O_CREAT|O_WRONLY, S_IREAD|S_IWRITE);

  if (_new_fd == -1)
    {
      throw ::strerror (errno);
    }

  {
    int status = ::dup2(_new_fd, _old_fd);
    int saved_errno = errno;

    ::close (_new_fd);

    if (-1 == status)
      {
        throw ::strerror (saved_errno);
      }
  }
}

~redirector ()
  {
    if (_fd_with_saved_info != -1)
      {
        if (-1 == ::dup2 (_fd_with_saved_info, _old_fd))
          perror ("dup2");

        if (-1 == ::close (_fd_with_saved_info))
          perror ("closing _fd_with_saved_info");
      }
  }
};

}
int main ()
{
  bool f_all_tests_passed = true;
  redirector shaddap (STDOUT_FILENO, "stdout.txt");

  TextTestResult SeatResult;
  seat_test::suite ()->run (&SeatResult);
  std::clog << SeatResult << std::endl;
  f_all_tests_passed = f_all_tests_passed && SeatResult.wasSuccessful ();

  TextTestResult CardResult;
  CardTest::suite ()->run (&CardResult);
  std::clog << CardResult << std::endl;
  f_all_tests_passed = f_all_tests_passed && CardResult.wasSuccessful ();

  TextTestResult AuctionResult;
  auction_test::suite ()->run (&AuctionResult);
  std::clog << AuctionResult << std::endl;
  f_all_tests_passed = f_all_tests_passed && AuctionResult.wasSuccessful ();

  TextTestResult ScoreResult;
  ScoreTest::suite ()->run (&ScoreResult);
  std::clog << ScoreResult << std::endl;
  f_all_tests_passed = f_all_tests_passed && ScoreResult.wasSuccessful ();

  return (f_all_tests_passed ? EXIT_SUCCESS : EXIT_FAILURE);
}
