// Copyright (C) 2001 Eric Hanchrow
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Eric Hanchrow -- offby1@blarg.net
// 1914 N. 40th
// Seatlle, WA, USA
// 98103-8310

#if defined (WIN32)
# pragma warning(disable: 4786)
#endif

#include <cctype>
#include <strstream>
#include <iostream>
#include <algorithm>
#include "text_player.hh"
#include "../auction.hh"
#include "../hand.hh"
#include "../score.hh"

namespace {

  // Return the "words" of the input string, where "words" are simply
  // non-space characters.

std::vector<std::string>
split (const std::string &input)
{
  std::string s (input);
  std::vector<std::string> return_value;

  do
    {
      while (s.size ()
             &&
             (isspace (s[0])))
        s.erase (s.begin ());

      if (s.size ())
      {
        std::string tmp;
        while (s.size ()
               &&
               (!isspace (s[0])))
          {
            tmp.append (1, s[0]);
            s.erase (s.begin ());
          }

        return_value.push_back (tmp);
      }
    } while (s.size ());

  return return_value;
}

};

#if defined (WIN32)
# define MIN std::_MIN
#else
# define MIN std::min
#endif

call *
text_player::make_call (const seat &who_am_i, // only for diagnostics and ruminations
                        const auction &so_far,
                        const rubber_scoresheet &score)
{
  // Display the calls leading up to this.
  {
    unsigned short calls_to_display = MIN (static_cast<size_t>(3), so_far.size ());

    std::vector<const call *>::const_iterator i = so_far.calls().end () - calls_to_display;

    if (i != so_far.calls().end ())
      {
        std::cout << "Here's the last few calls: " << std::endl;
      }

    for (;
         i != so_far.calls().end ();
         i++)
      {
        (*i)->print();
        std::cout << std::endl;
      }
  }

  std::cout << "Here is your hand: "
            << std::endl;
  _hand.print (std::cout, true) ;

  std:: cout << std::endl
             << "Please bid: ";

  call *return_value = 0;

  do
    {
      std::vector<std::string> words;

      {
        std::string line;
        std::getline (std::cin, line);

        words = split (line);
        if (!words.size ())
          {
            std::cerr << "I didn't hear you."
                      << std::endl;
            continue;
          }
      }

      if (words.size () == 1)
        {
          if (words[0] == "pass")
            {
              // hmm.  I wonder if there's a way I can ask the `pass'
              // class how it spells itself; that way I wouldn't have to
              // hard-code the string "pass" here.  And similarly for the
              // other calls, and for cards.
              return new pass;
            }

          else if (words[0] == "double")
            {
              return new Double;
            }

          else if (words[0] == "redouble")
            {
              return new Double (true);
            }
          else
            {
              std::cerr << "You uttered one word, `"
                        << words[0]
                        << "', but the only single words I understand are `pass',"
                        << " `double', and `redouble'."
                        << std::endl;
              continue;
            }
        }
      else if (words.size () == 2)
        {
          std::istrstream number (words[0].c_str());
          denomination d;
          int level = 0;
          number >> level;
          if (!number.good ())
            {
              std::cerr << "`"
                        << words[0]
                        << "' ain't no number!"
                        << std::endl;
              continue;
            }

          if (0 == std::string ("spades").find (words[1]))
            d = spades;
          else if (0 == std::string ("hearts").find (words[1]))
            d = hearts;
          else if (0 == std::string ("diamonds").find (words[1]))
            d = diamonds;
          else if (0 == std::string ("clubs").find (words[1]))
            d = clubs;
          else if (0 == std::string ("notrump").find (words[1]))
            ;                       // d defaults to notrump
          else
            {
              std::cerr << "That denomination stinks."
                        << std::endl;
              continue;
            }

          return_value = new bid (level, d);
        }
      else
        {
          std::cerr << "You talk too much."
                    << std::endl;
        }
    } while (!return_value);

  return return_value;
}

card
text_player::choose_card (const auction &,
                          const rubber_scoresheet &,
                          const plays &,
                          const std::vector<card> &current_trick,
                          const hand *dummy // will be 0 for opening lead
                          )
{
  Assert<internal_error> (_hand.size () > 0, "Can't choose a card when my hand is empty");

  // Display the current trick.
  {
    unsigned short plays_to_display = MIN (static_cast<size_t>(3), current_trick.size ());
    std::vector<card>::const_iterator i = current_trick.end () - plays_to_display;

    if (i != current_trick.end ())
      {
        std::cout << "Here's the current trick: " << std::endl;
      }
    else
      {
        std::cout << "Your lead: " << std::endl;
      }

    for (;
         i != current_trick.end ();
         i++)
      {
        std::cout << *i
                  << std::endl;
      }
  }

  if (dummy)
    {
      std::cout << "Dummy:" << std::endl;
      dummy->print (std::cout, false);
      std::cout << std::endl;
    }

  std::cout << "Your hand: " << std::endl;
  _hand.print (std::cout);
  std::cout << std::endl;
  hand::iterator choice = _hand.end();

  do
    {
      std::vector<std::string> words;
      std::cout << "Pick a card, any card: ";

      {
        std::string line;
        std::getline (std::cin, line);

        words = split (line);
      }

      if (!words.size ())
        {
          std::cerr << "I didn't hear you."
                    << std::endl;
          continue;
        }

      std::istrstream number (words[0].c_str());

      {
        unsigned int index = 0;
        number >> index;
        if (!number.good ()
            ||
            index >= _hand.size())
          {
            std::cerr << "That number stinks."
                      << std::endl;
            continue;
          }

        choice = _hand.begin () + index;
      }

    } while (choice == _hand.end());

  card return_value = *choice;
  _hand.erase (choice);
  return return_value;
}
